/*
 * Copyright (c) 2016-2017,2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Tue Aug 23 2016
 */
#include "aclibmrsim.h"

#include <math.h>
#include <stddef.h>

#include "private.h"

/* --- mrsim_motor_pwm ----------------------------------------------------- */

void
mrsim_motor_pwm(struct mrsim_motor_s *motor, const double pwm,
                const struct timespec *ts)
{
  if (motor->state.disabled) return;
  if (!motor->state.spinning) return;
  motor->wdts = *ts;

  if (motor->servo) {
    motor->servo = false;
    motor->state.wd = 0;
    motor->smc_bias = 0;
    motor->smc_gain = 0;
    motor->smc_err = 0;
  }
  motor->state.pwm = pwm;
}


/* --- mrsim_motor_period -------------------------------------------------- */

void
mrsim_motor_period(struct mrsim_motor_s *motor, const double wd,
                   const struct timespec *ts)
{
  if (motor->state.disabled) return;
  if (!motor->state.spinning) return;
  motor->wdts = *ts;

  if (!motor->servo) {
    motor->servo = true;
    motor->smc_bias = motor->state.pwm;
    motor->smc_gain = 1;
    motor->smc_err = 0;
  }
  motor->state.wd = wd;
}


/* --- mrsim_motor_control ------------------------------------------------- */

void
mrsim_motor_control(const struct mrsim_param_s *p,
                    struct mrsim_motor_s *motor)
{
  if (motor->state.starting) {
    /* startup PWM for 20Hz */
    motor->state.pwm =
      2*M_PI *
      (p->motor.K + p->motor.R * p->motor.b / p->motor.K) / p->motor.V * 20;
    if (motor->state.w > 15.) {
      motor->state.spinning = true;
      motor->state.starting = false;
      mrsim_motor_period(motor, 1e6/65535, &motor->wdts);
    }
    return;
  }

  if (!motor->state.spinning) {
    motor->state.pwm = 0.;
    return;
  }

  if (motor->servo) {
    double w;
    int16_t u;
    uint16_t t, p;

    w = motor->state.w + mrsim_gnoise(2);
    p = fabs(w) < 1000000./42/65535 ? 65535/42 : 1000000./42/fabs(w);

    t = fabs(motor->state.wd) < 1000000./65535 ?
      65535/42 : 1000000./42/fabs(motor->state.wd);

    /* error sign */
    if (t < p)
      motor->smc_err = (3 * motor->smc_err - 65536) / 4;
    else if (t > p)
      motor->smc_err = (3 * motor->smc_err + 65536) / 4;
    else
      motor->smc_err = 3 * motor->smc_err / 4;

    /* gain adaptation */
    if (motor->smc_err < -50000 || motor->smc_err > 50000) {
      if (motor->smc_gain < 512*motor->state.pwm) motor->smc_gain += 4;
    } else if (motor->smc_err < -32768 || motor->smc_err > 32768) {
      if (motor->smc_gain < 512*motor->state.pwm) motor->smc_gain += 2;
    } else {
      motor->smc_gain -= 2;
      if (motor->smc_gain < 1) motor->smc_gain = 1;
    }

    /* bias adaptation */
    if (motor->smc_err < -65000) {
      motor->smc_bias += 6;
      if (motor->smc_bias > 1023) motor->smc_bias = 1023;
    } else if (motor->smc_err < -50000) {
      if (motor->smc_bias < 1023) motor->smc_bias += 1;
    } else if (motor->smc_err > 65000) {
      motor->smc_bias -= 6;
      if (motor->smc_bias < 1) motor->smc_bias = 1;
    } else if (motor->smc_err > 50000) {
      if (motor->smc_bias > 1) motor->smc_bias -= 1;
    }

    /* command */
    if (t < p)
      u = motor->smc_bias + motor->smc_gain;
    else if (t > p)
      u = motor->smc_bias - motor->smc_gain;
    else
      u = motor->smc_bias;

    if (u > 1023) u = 1023;
    else if (u < 0) u = 0;

    motor->state.pwm =  u/1024.;
    return;
  }
}


/* --- mrsim_motor_model --------------------------------------------------- */

static void
mrsim_motor_eq(const struct mrsim_param_s *p,
               const double pwm, const double w, const double i,
               double *dw, double *di)
{
  *dw =
    - p->motor.b * w
    + p->motor.K * i / (2*M_PI);
  *di =
    - 2*M_PI * p->motor.K * w
    - p->motor.R * i
    + pwm * p->motor.V;
}


void
mrsim_motor_model(const struct mrsim_param_s *p, struct mrsim_motor_s *motor,
                  const double dt)
{
  struct { double dw, di; } k[4];

  /* RK4 */
  mrsim_motor_eq(p, motor->state.pwm,
                 motor->state.w,
                 motor->state.i,
                 &k[0].dw, &k[0].di);
  mrsim_motor_eq(p, motor->state.pwm,
                 motor->state.w + dt/p->motor.J/2 * k[0].dw,
                 motor->state.i + dt/p->motor.L/2 * k[0].di,
                 &k[1].dw, &k[1].di);
  mrsim_motor_eq(p, motor->state.pwm,
                 motor->state.w + dt/p->motor.J/2 * k[1].dw,
                 motor->state.i + dt/p->motor.L/2 * k[1].di,
                 &k[2].dw, &k[2].di);
  mrsim_motor_eq(p, motor->state.pwm,
                 motor->state.w + dt/p->motor.J * k[2].dw,
                 motor->state.i + dt/p->motor.L * k[2].di,
                 &k[3].dw, &k[3].di);

  motor->state.w +=
    dt/p->motor.J/6 * (k[0].dw + 2*k[1].dw + 2*k[2].dw + k[3].dw);
  motor->state.i +=
    dt/p->motor.L/6 * (k[0].di + 2*k[1].di + 2*k[2].di + k[3].di);

  if (motor->state.i > 40.) motor->state.i = 40.;
}


/* --- mrsim_update_motors ------------------------------------------------- */

void
mrsim_update_motors(const struct mrsim_s *mrsim)
{
  struct mrsim_motor_s *motor;
  uint8_t i, k, niter;

  motor = mrsim->motor;
  for(i = 0; i < mrsim->params->body.rotors; i++, motor++) {
    if (motor->state.disabled) continue;

    /* watchdog */
    if (mrsim->ts->now.tv_sec - motor->wdts.tv_sec
        + 1e-9 * (mrsim->ts->now.tv_nsec - motor->wdts.tv_nsec) > 0.5)
      mrsim_motor_period(motor, 1e6/65535, &mrsim->ts->now);

    /* emulate the variable period of the BL control algorithm */
    if (fabs(motor->state.w) < 25)
      niter = 1 + mrsim->ts->dt*1000;
    else
      niter = 1 + mrsim->ts->dt * 42*fabs(motor->state.w);

    for(k = 0; k < niter; k++) {
      mrsim_motor_control(mrsim->params, motor);
      mrsim_motor_model(mrsim->params, motor, mrsim->ts->dt/niter);
    }
  }
}


/* --- mrsim_motor_measure ------------------------------------------------- */

double
mrsim_motor_measure(const struct mrsimstate_motor_s *motor)
{
  uint16_t p;

  p = fabs(motor->w) < 1000000./65535 ? 65535 : 1000000./fabs(motor->w);
  p = p/42 + mrsim_gnoise(2);

  return 1000000./42/p;
}


/* --- mrsim_seqid --------------------------------------------------------- */

uint8_t
mrsim_seqid(struct timespec tv, double period)
{
  double t = tv.tv_sec - 1400000000; /* offset for better double precision */
  t += tv.tv_nsec * 1e-9;

  return fmod(t/period, 256);
}
