/*
 * Copyright (c) 2018-2021 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                           Anthony Mallet on Thu Mar 14 2019
 */
#include "aclibmrsim.h"

#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "private.h"

/* --- mrsim_init ---------------------------------------------------------- */

struct mrsim_s *
mrsim_init(uint8_t rotors)
{
  struct mrsim_s *mrsim;

  const struct mrsim_s defaults = {
    .pty = malloc(sizeof(*mrsim->pty)),
    .params = malloc(sizeof(*mrsim->params)),
    .motor = malloc(rotors * sizeof(*mrsim->motor)),
    .wrench = malloc(sizeof(*mrsim->wrench)),
    .body = malloc(sizeof(*mrsim->body)),
    .imu = malloc(sizeof(*mrsim->imu)),
    .ts = malloc(sizeof(*mrsim->ts))
  };
  const struct mrsim_param_s params_defaults = {
    .motor = {
      .J = 8e-5,
      .b = 5e-4,
      .K = 0.01,
      .R = 0.2,
      .L = 2e-4,
      .V = 16.
    },

    .body = {
      .rotors = rotors,
      .G = malloc(6 * rotors * sizeof(double)),
      .J = {
        0.015,    0.,    0.,
        0.,    0.015,    0.,
        0.,       0., 0.015
      },
      .m = 1.0,
      .g = 9.81,
      .B = { 23.816e-6, -0.410e-6, -39.829e-6 } /* in Toulouse, France (NWU) */
    },

    .state = {
      .pstddev = 1e-3, .qstddev = 1e-3,
      .vstddev = 1e-3, .wstddev = 3e-3,
      .astddev = 2e-2, .Bstddev = 8e-8,

      .imup = 0., .magp = 0., .batp = 0., .motp = 0.
    }
  };
  const struct mrsim_motor_s motor_defaults = {
    .state = {
      .emerg = false, .spinning = false, .starting = false, .disabled = false,
      .pwm = 0., .wd = 0., .w = 0., .i = 0.
    }
  };
  const double kf = 6.5e-4;
  const double c = 0.0154;
  const double d = 0.23;
  const double G_default[6][4] = {
    {    0.,   0.,    0.,    0. },
    {    0.,   0.,    0.,    0. },
    {    kf,   kf,    kf,    kf },

    {    0.,  d*kf,   0., -d*kf },
    { -d*kf,    0., d*kf,    0. },
    {  c*kf, -c*kf, c*kf, -c*kf }
  };

  uint8_t i, j;

  mrsim = calloc(1, sizeof(*mrsim));
  if (!mrsim) return NULL;

  memcpy(mrsim, &defaults, sizeof(*mrsim));
  if (!mrsim->pty) goto err;
  if (!mrsim->params) goto err;
  if (!mrsim->motor) goto err;
  if (!mrsim->wrench) goto err;
  if (!mrsim->body) goto err;
  if (!mrsim->imu) goto err;
  if (!mrsim->ts) goto err;

  /* defaults */
  mrsim->pty->fd = -1;

  memcpy(mrsim->params, &params_defaults, sizeof(*mrsim->params));
  if (!mrsim->params->body.G) goto err;
  for(i = 0; i < 6; i++)
    for(j = 0; j < rotors; j++)
      mrsim->params->body.G[i*rotors + j] = j < 4 ? G_default[i][j] : 0.;

  for(i = 0; i < rotors; i++)
    mrsim->motor[i] = motor_defaults;

  *mrsim->body = (struct mrsimstate_body_s){.px = 0. /* all zeroes */};
  *mrsim->ts = (struct mrsim_ts_s){.dt = 0.  /* all zeroes */};
  mrsim_update_imu(mrsim);

  return mrsim;

err:
  mrsim_fini(mrsim);
  return NULL;
}


/* --- mrsim_fini ---------------------------------------------------------- */

void
mrsim_fini(struct mrsim_s *mrsim)
{
  if (!mrsim) return;

  if (mrsim->pty) {
    mrsim_close_pty(mrsim->pty);
    free(mrsim->pty);
  }
  if (mrsim->params) {
    if (mrsim->params->body.G) free(mrsim->params->body.G);
    free(mrsim->params);
  }
  if (mrsim->motor) free(mrsim->motor);
  if (mrsim->body) free(mrsim->body);
  if (mrsim->imu) free(mrsim->imu);
  if (mrsim->ts) free(mrsim->ts);
  free(mrsim);
}


/* --- mrsim_params -------------------------------------------------------- */

struct mrsim_param_s *
mrsim_params(const struct mrsim_s *mrsim)
{
  return mrsim->params;
}


/* --- mrsim_motors -------------------------------------------------------- */

struct mrsimstate_motor_s *
mrsim_motor(const struct mrsim_s *mrsim, uint8_t index)
{
  return &mrsim->motor[index < mrsim->params->body.rotors ? index : 0].state;
}


/* --- mrsim_wrench -------------------------------------------------------- */

struct mrsimstate_wrench_s *
mrsim_wrench(const struct mrsim_s *mrsim)
{
  return mrsim->wrench;
}


/* --- mrsim_wrench -------------------------------------------------------- */

struct mrsimstate_body_s *
mrsim_body(const struct mrsim_s *mrsim)
{
  return mrsim->body;
}


/* --- mrsim_update_time --------------------------------------------------- */

void
mrsim_update_time(const struct mrsim_s *mrsim, struct timespec now)
{
  struct mrsim_s *m = (struct mrsim_s *)mrsim;

  m->ts->dt =
    (now.tv_sec - m->ts->now.tv_sec)
    + 1e-9 * (now.tv_nsec - m->ts->now.tv_nsec);
  if (m->ts->dt < 1e-5) m->ts->dt = 1e-5;

  m->ts->now = now;
}


/* --- mrsim_process_msgs -------------------------------------------------- */

int
mrsim_process_msgs(const struct mrsim_s *mrsim)
{
  uint8_t *msg;
  ssize_t len;
  int e;

  while(1) {
    /* read incoming data */
    len = mrsim_recv_msg(mrsim->pty, &msg);
    if (len < 0) {
      switch(e = errno) {
        case EAGAIN: /* no data */
        case EIO: /* not connected */
          return 0;

        default:
          warn("receiving from %s", mrsim->pty->device);
          return e;
      }
    }
    if (len == 0) return 0;

    /* handle incoming messages */
    switch(msg[0]) {
      case 'i': /* IMU data */
        if (len == 5) {
          uint32_t p =
            ((uint32_t)msg[1] << 24) |
            ((uint32_t)msg[2] << 16) |
            ((uint32_t)msg[3] << 8) |
            ((uint32_t)msg[4] << 0);

          mrsim->params->state.imup = p/1e6;
        }
        break;

      case 'c': /* magnetometer data */
        if (len == 5) {
          uint32_t p =
            ((uint32_t)msg[1] << 24) |
            ((uint32_t)msg[2] << 16) |
            ((uint32_t)msg[3] << 8) |
            ((uint32_t)msg[4] << 0);

          mrsim->params->state.magp = p/1e6;
        }
        break;

      case 'b': /* battery level */
        if (len == 5) {
          uint32_t p =
            ((uint32_t)msg[1] << 24) |
            ((uint32_t)msg[2] << 16) |
            ((uint32_t)msg[3] << 8) |
            ((uint32_t)msg[4] << 0);

          mrsim->params->state.batp = p/1e6;
        }
        break;

      case 'm': /* motor data */
        if (len == 5) {
          uint32_t p =
            ((uint32_t)msg[1] << 24) |
            ((uint32_t)msg[2] << 16) |
            ((uint32_t)msg[3] << 8) |
            ((uint32_t)msg[4] << 0);

          mrsim->params->state.motp = p/1e6;
        }
        break;

      case 'g': /* start motors */ {
        unsigned int i;

        for(i = 0; i < mrsim->params->body.rotors; i++) {
          if (len > 1 && msg[1] != i + 1) continue;
          if (mrsim->motor[i].state.disabled) continue;
          if (mrsim->motor[i].state.starting) continue;
          if (mrsim->motor[i].state.spinning) continue;

          mrsim->motor[i].state.starting = true;
          mrsim->motor[i].state.spinning = false;
          mrsim->motor[i].servo = false;
        }
        break;
      }

      case 'x': /* stop motors */ {
        unsigned int i;

        for(i = 0; i < mrsim->params->body.rotors; i++) {
          if (len > 1 && msg[1] != i + 1) continue;
          if (mrsim->motor[i].state.disabled) continue;

          mrsim->motor[i].state.starting = false;
          mrsim->motor[i].state.spinning = false;
          mrsim->motor[i].servo = false;
        }
        break;
      }

      case 'w': /* velocity array */ {
        unsigned int i;
        int32_t p;

        for(i = 0; i < mrsim->params->body.rotors; i++) {
          if (len < 3 + 2 * i) break;

          p =
            ((int32_t)msg[1 + 2 * i] << 9) |
            ((uint32_t)msg[2 + 2 * i] << 1);
          mrsim_motor_period(mrsim->motor + i, 1e6/p, &mrsim->ts->now);
        }
        break;
      }

      case 'q': /* pwm array */ {
        unsigned int i;
        int32_t p;

        for(i = 0; i < mrsim->params->body.rotors; i++) {
          if (len < 3 + 2 * i) break;

          p =
            ((int32_t)msg[1 + 2 * i] << 9) |
            ((uint32_t)msg[2 + 2 * i] << 1);
          mrsim_motor_pwm(mrsim->motor + i, p/1023., &mrsim->ts->now);
        }
        break;
      }

      case '?': /* id */
        if (len == 1) mrsim_send_msg(mrsim->pty, "?chimera1.1");
        break;
    }
  }

  return 0;
}


/* --- mrsim_send_msgs ----------------------------------------------------- */

int
mrsim_send_msgs(const struct mrsim_s *mrsim)
{
  uint8_t i;

  /* send IMU data */
  if (mrsim->params->state.imup > 0 &&
      mrsim->ts->now.tv_sec - mrsim->ts->imu.tv_sec
      + 1e-9 * (mrsim->ts->now.tv_nsec - mrsim->ts->imu.tv_nsec) >
      0.95 * mrsim->params->state.imup) {
    mrsim->ts->imu = mrsim->ts->now;
    mrsim_send_msg(
      mrsim->pty, "I%1%2%2%2%2%2%2",
      mrsim_seqid(mrsim->ts->now, mrsim->params->state.imup),
      (int16_t)(32768/(8*9.81) * mrsim->imu->a[0]),
      (int16_t)(32768/(8*9.81) * mrsim->imu->a[1]),
      (int16_t)(32768/(8*9.81) * mrsim->imu->a[2]),
      (int16_t)(32768/(1000.*M_PI/180) * mrsim->imu->w[0]),
      (int16_t)(32768/(1000.*M_PI/180) * mrsim->imu->w[1]),
      (int16_t)(32768/(1000.*M_PI/180) * mrsim->imu->w[2]));
  }

  /* send magnetometer data */
  if (mrsim->params->state.magp > 0 &&
      mrsim->ts->now.tv_sec - mrsim->ts->mag.tv_sec
      + 1e-9 * (mrsim->ts->now.tv_nsec - mrsim->ts->mag.tv_nsec) >
      0.95 * mrsim->params->state.magp) {
    mrsim->ts->mag = mrsim->ts->now;
    mrsim_send_msg(
      mrsim->pty, "C%1%2%2%2",
      mrsim_seqid(mrsim->ts->now, mrsim->params->state.magp),
      (int16_t)(1e8 * mrsim->imu->B[0]),
      (int16_t)(1e8 * mrsim->imu->B[1]),
      (int16_t)(1e8 * mrsim->imu->B[2]));
  }

  /* send battery data */
  if (mrsim->params->state.batp > 0 &&
      mrsim->ts->now.tv_sec - mrsim->ts->bat.tv_sec
      + 1e-9 * (mrsim->ts->now.tv_nsec - mrsim->ts->bat.tv_nsec) >
      0.9 * mrsim->params->state.batp) {
    mrsim->ts->bat = mrsim->ts->now;
    mrsim_send_msg(
      mrsim->pty, "B%1%2",
      mrsim_seqid(mrsim->ts->now, mrsim->params->state.batp),
      (uint16_t)(1000.*mrsim->params->motor.V));
  }

  /* send motor data */
  if (mrsim->params->state.motp > 0 &&
      mrsim->ts->now.tv_sec - mrsim->ts->mot.tv_sec
      + 1e-9 * (mrsim->ts->now.tv_nsec - mrsim->ts->mot.tv_nsec) >
      0.95 * mrsim->params->state.motp) {
    mrsim->ts->mot = mrsim->ts->now;

    for(i = 0; i < mrsim->params->body.rotors; i++) {
      if (mrsim->motor[i].state.disabled) continue;

      double w = mrsim_motor_measure(&mrsim->motor[i].state);
      int16_t p;
      uint8_t s;

      s = i + 1;
      if (mrsim->motor[i].state.emerg) s |= 0x80;
      if (mrsim->motor[i].state.spinning) s |= 0x20;
      if (mrsim->motor[i].state.starting) s |= 0x10;

      p = (fabs(w) > 1e6/65535) ? 1e6/2/w : copysign(32767, w);

      mrsim_send_msg(
        mrsim->pty, "M%1%1%2%2%2",
        mrsim_seqid(mrsim->ts->now, mrsim->params->state.motp),
        s, p,
        (int16_t)(1024.*copysign(mrsim->motor[i].state.pwm, w)),
        (uint16_t)(1000.*mrsim->motor[i].state.i));
    }
  }

  return 0;
}
